/*
 *
 * CLEX File Manager
 *
 * Copyright (C) 1997-2000 Tempest s.r.o. http://www.tempest.sk
 * Copyright (C) 2001-2002 Vlado Potisk <vlado_potisk@clex.sk>
 *
 * CLEX is free software without warranty of any kind; see the
 * GNU General Public License as set out in the "COPYING" document
 * which accompanies the CLEX File Manager package.
 *
 * CLEX can be downloaded from http://www.clex.sk
 *
 */

#include <autoconf.h>

#include <sys/types.h>		/* clex.h */
#include <ctype.h>			/* iscntrl() */
#include <stdarg.h>			/* va_list */
#include <string.h>			/* strcpy() */
#ifdef HAVE_NCURSES_H
# include <ncurses.h>		/* initscr() */
#else
# include <curses.h>
#endif
#include <term.h>			/* enter_bold_mode */

#include "clex.h"
#include "inout.h"

#include "cfg.h"			/* config_num() */
#include "completion.h"		/* compl_typestr() */
#include "control.h"		/* get_current_mode() */
#include "directory.h"		/* dir_split_len() */
#include "edit.h"			/* edit_adjust() */
#include "err.h"			/* err_exit() */
#include "panel.h"			/* pan_adjust() */
#include "tty.h"			/* tty_press_enter() */
#include "userdata.h"		/* get_mylogin_at_host() */
#include "ustring.h"		/* USTR() */

#ifndef A_NORMAL
# define A_NORMAL 0
#endif

#ifndef A_BOLD
# define A_BOLD A_STANDOUT
#endif

#ifndef A_REVERSE
# define A_REVERSE A_STANDOUT
#endif

#ifndef A_UNDERLINE
# define A_UNDERLINE A_STANDOUT
#endif

#ifndef ACS_HLINE
# define ACS_HLINE '-'
#endif

/* minimal required screen size */
#define MIN_COLS	64
#define MIN_LINES	12

/* safe addch() */
#define ADDCH(X)	(iscntrl(X) ? addch('?' | attrb) : addch(X))

/* win_position() messages */
#define MSG_NORMAL	0
#define MSG_WAIT	1
#define MSG_NODATA	2

static const char *remark = 0;	/* this remark is on the screen */
static const char *warnmsg = 0;	/* this warning is on the screen */
static const char *layout_panel;/* file panel layout - see CFG_LAYOUT */
static const char *layout_line;	/* info line layout */
static FLAG waitmsg = 0;		/* "PLEASE WAIT" should disappear */
static int prevkey = 0;			/* previously pressed key */
static chtype attrr, attrb;		/* reverse and bold or substitutes */
static void (*put_line)(int);	/* function drawing panel lines */
static int framechar;			/* frame character from win_frame() */

static void win_info(void);		/* defined below */
static void win_position(int);	/* defined below */

/* draw everything from scratch */
static void
screen_draw_all(void)
{
	int y, x;

	for (;/* until break */;) {
		clear();
		getmaxyx(stdscr,y,x);
		display.scrcols = x;
		display.panlines = y - config_num(CFG_CMD_LINES) - 5;
		/*
		 * there are 2 reserved positions: the bottom-right corner
		 * is always left untouched to prevent automatic scrolling
		 * and the position just before it can be occupied only
		 * either by the cursor or by the '>' continuation mark
		 */
		display.textline_area = x * config_num(CFG_CMD_LINES) - 2;
		if (x >= MIN_COLS && y >= MIN_LINES)
			break;
		/* window too small */
		printw("SCREEN: this window %d x %d is too small, "
		  "required is %d x %d or larger; press ctrl-C to "
		  "exit CLEX or (if possible) enlarge the window"
#ifndef KEY_RESIZE
		  " and press any other key to continue"
#endif
		  ". ",y,x,MIN_LINES,MIN_COLS);
		refresh();
		if (getch() == CH_CTRL('C'))
			err_exit("Window too small");
	}
	win_heading();
	win_frame();
	win_bar();
	if (panel) {
		/* panel is NULL only when starting clex */
		pan_adjust(panel);
		win_panel();
		win_info();
	}
	if (textline) {
		edit_adjust();
		win_edit();
	}
}

/* start CURSES */
void
curses_initialize(void)
{
	win_layout_reconfig();
	if (TCLR(display.wait))
		tty_press_enter();

	initscr();			/* restores signal dispositions on FreeBSD ! */
	signal_initialize();	/* FreeBSD initscr() bug workaround ! */
	raw();
	nonl();
	noecho();
	keypad(stdscr,TRUE);
#ifdef HAVE_NOTIMEOUT
	notimeout(stdscr,TRUE);
#endif
	clear();
	refresh();
	display.curses = 1;

	if (enter_reverse_mode && *enter_reverse_mode)
		attrr = A_REVERSE;
	else
		attrr = A_STANDOUT;
	if (enter_bold_mode && *enter_bold_mode)
		attrb = A_BOLD;
	else if (enter_underline_mode && *enter_underline_mode)
		attrb = A_UNDERLINE;
	else
		attrb = A_STANDOUT;

	screen_draw_all();
}

/* restart CURSES */
void
curses_restart(void)
{
	if (TCLR(display.wait))
		tty_press_enter();

	reset_prog_mode();
	touchwin(stdscr);
	display.curses = 1;
	screen_draw_all();
}

/* stop CURSES */
void
curses_stop(void)
{
	clear();
	refresh();
	endwin();
	display.curses = 0;
}

/* set cursor to the proper position and refresh screen */
static void
screen_refresh(void)
{
	int pos;

	if (waitmsg)
		win_position(MSG_NORMAL);

	if (textline == 0)
		pos = 0;
	else if (textline->offset)
		pos = 1 + textline->curs - textline->offset;
	else
		pos = textline->curs + textline->promptlen;
	move(display.panlines + 5 + pos / display.scrcols,
	  pos % display.scrcols);
	refresh();
}

/*
 * like printf() but curses mode is suspended if necessary,
 * txt_printf() is to be used in xxx_reconfig() functions
 * which may be invoked in both curses and non-curses modes.
 */
void
txt_printf(const char *format, ...)
{
	va_list argptr;

	display.wait = 1;
	if (display.curses) {
		curses_stop();
		puts("\nReconfiguring CLEX");
	}

	va_start(argptr,format);
	vprintf(format,argptr);
	va_end(argptr);
	fflush(stdout);
}

/****** keyboard input functions ******/

/*
 * get next input char, no processing except screen resize event,
 * use this input function to get unfiltered input
 */
int
kbd_getraw(void)
{
	int key;

	for (;/* until return */;) {
		screen_refresh();
		do {
			key = getch();
		} while (key == -1);	/* where is this -1 coming from ?! */
#ifdef KEY_RESIZE
		if (key == KEY_RESIZE) {
			screen_draw_all();
			continue;
		}
#endif
		return key;
	}
}

/*
 * get next input char, not saved for kbd_esc() usage,
 * use this function to get a single key unrelated to the rest
 * of input stream (e.g. "press any key to continue")
 */
static int
kbd_getany(void)
{
	int key;

	while ((key = kbd_getraw()) == CH_CTRL('L'))
		/* redraw screen */
		wrefresh(curscr);
	return key;
}

/* get next input char */
int
kbd_input(void)
{
	static int key = 0;

	prevkey = key;
	key = kbd_getany();	
	/* <esc> 1 --> <F1>, <esc> 2 --> <F2>, ... <esc> 0 --> <F10> */
	if (prevkey == CH_ESC && isdigit(key))
		key = KEY_F((key - '0' + 9) % 10 + 1);

	/* clean remark (if any) */
	if (remark) {
		remark = 0;
		win_info();
	}

	return key;
}

/* was the previous key an ESC ? */
int
kbd_esc(void)
{
	return prevkey == CH_ESC;
}

/****** output functions ******/

/*
 * following functions write to these screen areas:
 *
 * win_heading							/usr/local
 * win_frame							----------
 * win_panel							 DIR bin
 * win_panel							>DIR etc <
 * win_panel							 DIR man
 * win_frame, win_waitmsg, win_position	---------< 3/10 >-
 * win_info, win_warning, win_remark	0755 rwxr-xr-x
 * win_bar								[ CLEX file manager ]
 * win_edit								shell $ ls -l_
 */

void
win_layout_reconfig(void)
{
	static USTRING layout = { 0,0 };
	char *pch;

	us_copy(&layout,config_str(CFG_LAYOUT));

	/* split layout to panel fields and line fields */
	if ( (pch = strchr(layout_panel = USTR(layout),':')) ) {
		*pch = '\0';
		layout_line = pch + 1;
	}
	else {
		layout_line = "";	
		txt_printf("CONFIG: Incorrect \"LAYOUT\": missing colon ':'\n");
	}
}

/* write a line of repeating chars */
static void
char_line(int ch, int cnt)
{
	while (cnt-- > 0)
		addch(ch);
}

void
win_frame(void)
{
	switch(config_num(CFG_FRAME)) {
	case 0:
		framechar = '-';
		break;
	case 1:
		framechar = '=';
		break;
	default:
		framechar = ACS_HLINE;
	}

	move(1,0);
	char_line(framechar,display.scrcols);
	move(display.panlines + 2,0);
	char_line(framechar,display.scrcols);
}

static void
win_position(int msg)
{
	static int prev_len = 0;
	char *pos, buffer[40];
	int len;

	if (msg == MSG_WAIT) {
		pos = "< PLEASE WAIT >";
		len = 15;	/* strlen(pos) */
	}
	else {
		waitmsg = 0;
		if (msg == MSG_NODATA) {
			pos = "< NO DATA >";
			len = 11;
		}
		else if (panel->type == PANEL_TYPE_FILE &&
		  ppanel_file->selected)
			len = sprintf(pos = buffer,"< [%d] %d/%d >",
			  ppanel_file->selected,panel->curs + 1,panel->cnt);
		else
			len = sprintf(pos = buffer,"< %d/%d >",panel->curs + 1,
			  panel->cnt);
	}

	if (prev_len > len) {
		move(display.panlines + 2,display.scrcols - prev_len - 2);
		char_line(framechar,prev_len - len);
	}
	else
		move(display.panlines + 2,display.scrcols - len - 2);
	if (msg != MSG_NORMAL)
		attrset(A_BOLD);
	addstr(pos);
	if (msg != MSG_NORMAL)
		attrset(A_NORMAL);
	prev_len = len;
}

void
win_waitmsg(void)
{
	if (display.curses) {
		win_position(MSG_WAIT);
		screen_refresh();
	}
}

void
win_nowaitmsg(void)
{
	if (display.curses)
		waitmsg = 1;
}

/*
 * putstr_trunc() writes string 'str' padding or truncating it
 * to the total length of 'maxwidth'. Its behavior might be altered
 * by OPT_XXX options (may be OR-ed together).
 *
 * putstr_trunc() returns the total number of characters written
 * (only useful with OPT_NOPAD)
 */
#define OPT_NOPAD	1	/* do not pad */
#define OPT_NOCONT	2	/* do not write continuation mark '>' */
#define OPT_SQUEEZE	4	/* squeeze long str in the middle: abc...xyz */
static int
putstr_trunc(const char *str, int maxwidth, int options)
{
	int i, ch, len, dots, part1, part2;

	if (maxwidth <= 0)
		return 0;

	if (options & OPT_SQUEEZE) {
		len = strlen(str);
		if (len > maxwidth) {
			dots = maxwidth >= 6 ? 4 : 1;
			part2 = 5 * (maxwidth - dots) / 8;
			part1 = maxwidth - dots - part2;
			putstr_trunc(str,part1,OPT_NOCONT);
			char_line('.',dots);
			putstr_trunc(str + len - part2,part2,0);
			return maxwidth;
		}
		/* else: SQUEEZE option is superfluous --> ignored */
	}

	for (i = 0; i < maxwidth - 1; i++)
		if ( (ch = str[i] & 0xFF) )
			ADDCH(ch);
		else {
			/* EOL */
			if (options & OPT_NOPAD)
				return i;
			char_line(' ',maxwidth - i);
			return maxwidth;
		}

	ch = str[i] & 0xFF;
	if (ch == '\0')
		addch(' ');
	else if (str[i + 1] && !(options & OPT_NOCONT))
		addch('>' | attrb);	/* continuation mark in bold font */
	else
		ADDCH(ch);
	return maxwidth;
}

/* file panel heading: primary + secondary panel's directory */
static void
twodirs(void)
{
	int len1, len2, opt1, opt2, width;
	const char *dir1, *dir2;

	/*
	 * directory names are separated by 2 spaces
	 * their width is kept in ratio 5:3
	 */
	width = display.scrcols - 2;				/* available room */
	len1 = strlen(dir1 = USTR(ppanel_file->dir));
	len2 = strlen(dir2 = USTR(ppanel_file->other->dir));
	opt1 = opt2 = 0;
	if (len1 + len2 <= width)
		len1 = width - len2;					/* enough room */
	else if (len1 <= (5 * width) / 8) {
		len2 = width - len1;					/* squeeze second */
		opt2 = OPT_SQUEEZE;
	}
	else if (len2 <= (3 * width) / 8) {
		len1 = width - len2;					/* squeeze first */
		opt1 = OPT_SQUEEZE;
	}
	else {
		len1 = (5 * width) / 8;					/* squeeze both */
		len2 = width - len1;
		opt1 = opt2 = OPT_SQUEEZE;
	}
	attrset(attrb);
	putstr_trunc(dir1,len1,opt1);
	attrset(A_NORMAL);
	addstr("  ");
	putstr_trunc(dir2,len2,opt2);
}

/* panel heading - top screen line */
void
win_heading(void)
{
	int mode, len;
	const char *msg;

	move(0,0);

	mode = get_current_mode();
	if (mode == MODE_FILE_MAIN) {
		twodirs();
		return;
	}

	addch(' ');
	len = 1;
	switch (mode) {
	case MODE_COMPLETION:
		msg = compl_typestr(1);
		addstr(msg);
		len = 1 + strlen(msg);
		msg = "COMPLETION";
		break;
	case MODE_CONFIG_MAIN:
		msg = "CONFIGURATION: <enter> = change, U = undo, D = default";
		break;
	case MODE_CONFIG_EDIT_NUM:
	case MODE_CONFIG_EDIT_STR:
		msg = "CONFIGURATION: enter new value";
		break;
	case MODE_CONFIG_CONFIRM:
		msg = "CONFIGURATION: confirm changes";
		break;
	case MODE_DIR_EDIT:
		msg = "CHANGE WORKING DIRECTORY: enter new directory";
		break;
	case MODE_DIR_MAIN:
	case MODE_DIR_SPLIT:
		msg = "CHANGE WORKING DIRECTORY";
		break;
	case MODE_HELP:
		addstr("HELP: ");
		len = 7;
		attrset(attrb);
		msg = panel_help.heading;
		break;
	case MODE_HIST:
		msg = "COMMAND HISTORY: <tab> = insert, <enter> = execute";
		break;
	case MODE_LOCATE:
		msg = "FIND FILES: enter first character(s)";
		break;
	case MODE_MENU:
		msg = "MAIN FUNCTION MENU";
		break;
	case MODE_SELECT:
	case MODE_DESELECT:
		msg = "SELECT/DESELECT FILES: enter pattern";
		break;
	case MODE_SORT:
		msg = "SET FILE SORT ORDER";
		break;
	default:
		msg = "";
	}
	putstr_trunc(msg,display.scrcols - len,0);
	if (mode == MODE_HELP)
		attrset(A_NORMAL);
}

/*
 * see CFG_LAYOUT for more information about 'fields'
 * function returns the remaining unused width
 */
static int
print_fields(FILE_ENTRY *pfe, int width, const char *fields)
{
	char ch;
	const char *txt1, *txt2;
	int fw;
	static char *typestring[] = {
		"    ", "exec", "suid", "Suid", "sgid", "/DIR", "/MNT",
		"Bdev", "Cdev", "FIFO", "sock", "spec", "  ??"
	};	/* must match the FT_XXX #defines */

	while ( (ch = *fields++) ) {
		switch (ch) {
		case '*':	/* selection mark */
			fw = 2;
			txt1 = 0;
			txt2 = pfe->select ? "*" : " ";
			break;
		case 'a':	/* access date/time */
			fw = FE_TIME_STR;
			txt1 = 0;
			txt2 = pfe->atime_str;
			break;
		case 'A':	/* access date/time (alternative format) */
			fw = FE_TIME_STR + 4;
			txt1 = "acc=";
			txt2 = pfe->atime_str;
			break;
		case 'd':	/* modification date/time */
			fw = FE_TIME_STR;
			txt1 = 0;
			txt2 = pfe->mtime_str;
			break;
		case 'D':	/* modification date/time (alternative format) */
			fw = FE_TIME_STR + 4;
			txt1 = "mod=";
			txt2 = pfe->mtime_str;
			break;
		case 'i':	/* inode change date/time */
			fw = FE_TIME_STR;
			txt1 = 0;
			txt2 = pfe->ctime_str;
			break;
		case 'I':	/* inode date/time (alternative format) */
			fw = FE_TIME_STR + 4;
			txt1 = "ino=";
			txt2 = pfe->ctime_str;
			break;
		case 'l':	/* links */
			fw = FE_LINKS_STR;
			txt1 = 0;
			txt2 = pfe->links_str;
			break;
		case 'L':	/* links (alternative format) */
			fw = FE_LINKS_STR + 3;
			txt2 = pfe->links_str;
			txt1 = *txt2  == ' ' ?  "   " : "lnk";
			break;
		case 'm':
		case 'M':	/* mode (normal or alternative format) */
			fw = FE_MODE1_STR;
			txt1 = 0;
			txt2 = pfe->mode1_str;
			break;
		case 'o':	/* owner */
			fw = FE_OWNER_STR;
			txt1 = 0;
			txt2 = pfe->owner_str;
			break;
		case 'p':	/* permission */
			fw = FE_MODE2_STR;
			txt1 = 0;
			txt2 = pfe->mode2_str;
			break;
		case 's':	/* file size (device major/minor) */
			txt1 = 0;
			if (IS_FT_DEV(pfe->file_type)) {
				fw = FE_DEV_STR;
				txt2 = pfe->sd_str.device;
			}
			else {
				fw = FE_SIZE_STR;
				txt2 = pfe->sd_str.size;
			}
			break;
		case 't':	/* file type */
			fw = 7;
			txt1 = pfe->symlink ? "->" : "  ";
			txt2 = typestring[pfe->file_type];
			break;
		default:	/* syntax error */
			fw = 2;
			txt1 = 0;
			txt2 = "?";
		}

		if (width < fw)
			break;

		if (txt1)
			addstr(txt1);
		addstr(txt2);
		addch(' ');
		width -= fw;
	}
	return width;
}

/* information line */
static void
win_info(void)
{
	int width;
	const char *aux;
	FILE_ENTRY *pfe;

	move(display.panlines + 3,0);

	/* warnmsg has greater priority than a remark */
	if (warnmsg) {
		flash();
		attrset(attrb);
		putstr_trunc(warnmsg,display.scrcols - 16,OPT_NOPAD);
		attrset(A_NORMAL);
		addstr(" Press any key.");
		clrtoeol();
		return;
	}

	if (remark) {
		attrset(attrb);
		addstr("-- ");
		putstr_trunc(remark,display.scrcols - 7,OPT_NOPAD);
		addstr(" --");
		attrset(A_NORMAL);
		clrtoeol();
		return;
	}

	if (panel->cnt == 0 || panel->curs < 0) {
		clrtoeol();
		return;
	}

	if (panel->type == PANEL_TYPE_CONFIG) {
		addstr("  ");
		putstr_trunc(panel_config.config[panel->curs].help,
		  display.scrcols - 2,0);
		return;
	}

	if (panel->type == PANEL_TYPE_COMPL) {
		if ( (aux = panel_compl.candidate[panel->curs].aux) ) {
			addstr("  ");
			putstr_trunc(aux,display.scrcols - 2,0);
		}
		else
			clrtoeol();
		return;
	}

	if (panel->type == PANEL_TYPE_FILE)
		pfe = ppanel_file->files[panel->curs];
	else if (panel->type == PANEL_TYPE_LOCATE)
		pfe = panel_locate.files[panel->curs];
	else {
		clrtoeol();
		return;
	}

	if (pfe->file_type == FT_NA) {
		addstr("no status information available");
		clrtoeol();
	}
	else {
		width = print_fields(pfe,display.scrcols,layout_line);
		char_line(' ',width);
	}
}

/* warning message - can be used also in text mode */
void
win_warning(const char *msg)
{
	if (!display.curses) {
		putchar('\n');
		puts(msg);
		fflush(stdout);
		return;
	}

	/*
	 * win_info() is responsible for this screen area,
	 * real work is performed there
	 */
	warnmsg = msg;
	win_info();
	kbd_getany();
	warnmsg = 0;
	win_info();
}

void
win_remark(const char *format, ...)
{
	va_list argptr;
	static char buff[128];

	if (!display.curses) {
		txt_printf(
		  "BUG: win_remark(\"%s\", ...)\n"
		  "     called in text mode\n",format);
		return;
	}

	va_start(argptr,format);
	vsnprintf(buff,sizeof(buff),format,argptr);
	va_end(argptr);
	remark = buff;
	win_info();
}

void
win_ctrlchar(int ch)
{
	if (ch >= 1 && ch <= 26)
		win_remark("Ctrl-%c  ASCII %d",ch + 'A' - 1,ch);
	else
		win_remark("ASCII %d",ch);
}

void
win_completion(int cnt, const char *chars)
{
	int i;
	FLAG cntrl;
	char buffer[320], *dst;

	if (chars == 0) {
		win_remark("%d completion possibilities, "
		  "press <TAB> again for more info",cnt);
		return;
	}

	cntrl = 0;
	dst = buffer;
	for (i = 1; i < 256; i++)
		if (chars[i]) {
			if (iscntrl(i))
				cntrl = 1;
			else
				*dst++ = i;
		}
	strcpy(dst, chars[0] ? "<end>" : "");

	win_remark("next char: %s%s",
	  buffer, cntrl ? " (non-printables excluded)" : "");
}

void
win_bar(void)
{
	attrset(attrr);
	move(display.panlines + 4,0);
	putstr_trunc(adminmode ?
	  " CLEX file manager - ADMIN MODE" :
	  " CLEX file manager - press ctrl-G for main menu",
	  display.scrcols,0);
	move(display.panlines + 4,
	  display.scrcols - strlen(clex_user.login_at_host) - 2);
	addch(' ');
	addstr(clex_user.login_at_host);
	attrset(A_NORMAL);
}

void
win_edit(void)
{
	int width, offset;

	move(display.panlines + 5,0);

	/* special case: no textline */
	if (textline == 0) {
		clrtobot();
		return;
	}

	/* special case: empty line (prompt not highlighted) */
	if (textline->size == 0) {
		addstr(textline->prompt);
		clrtobot();
		return;
	}

	/* prompt (or continuation mark) */
	offset = textline->offset;
	attrset(attrb);
	width = display.textline_area + 1;
	if (offset == 0) {
		addstr(textline->prompt);
		width -= textline->promptlen;
	}
	else {
		addch('<');
		width--;
	}
	attrset(A_NORMAL);
	putstr_trunc(USTR(textline->line) + offset,width,0);
}

/****** win_panel() and friends  ******/

static void
line_pfe(FILE_ENTRY *pfe)
{
	int width;

	if (pfe->select)
		attron(attrb);

	/* 8 columns reserved for the filename */
	width = 8 + print_fields(pfe,display.scrcols - 10,layout_panel);
	if (pfe->symlink) {
		width -= putstr_trunc(USTR(pfe->file),width,OPT_NOPAD);
		if (width > 4) {
			addstr(" -> ");
			width -= 4;
			if (width > 0)
				putstr_trunc(USTR(pfe->link),width,0);
		}
		else
			char_line(' ',width);
	}
	else 
		putstr_trunc(USTR(pfe->file),width,0);

	if (pfe->select)
		attroff(attrb);
}

static void
line_file(int ln)
{
	line_pfe(ppanel_file->files[ln]);
}

static void
line_locate(int ln)
{
	line_pfe(panel_locate.files[ln]);
}

static void
line_dir_split(int ln)
{
	int width;

	addch(' ');
	width = dir_split_len(ln);
	if (width > display.scrcols - 3)
		putstr_trunc(panel_dir_split.dirname,display.scrcols - 3,0);
	else {
		putstr_trunc(panel_dir_split.dirname,width,OPT_NOCONT);
		char_line(' ',display.scrcols - 3 - width);
	}
}

static void
line_dir_browse(int ln)
{
	int shlen;

	addch(' ');
	if (ln <= TOPNDIRS + 1)
		putstr_trunc(panel_dir_browse.dir[ln].name,
		  display.scrcols - 3,0);
	else {
		shlen = panel_dir_browse.dir[ln].shlen;
		if (ln == panel_dir_browse.top || shlen >= display.scrcols - 5)
			shlen = 0;
		if (shlen) {
			char_line(' ',shlen - 2);
			addstr("__");
		}
		putstr_trunc(panel_dir_browse.dir[ln].name + shlen,
		  display.scrcols - shlen - 3,0);
	}
}

static void
line_hist(int ln)
{
	addch(' ');
	putstr_trunc(PUSTR(panel_hist.list[ln]),display.scrcols - 3,0);
}

static void
line_menu(int ln)
{
	/*
	 * there are two PANEL_MENU panels,
	 * must reference the data through the 'panel' pointer
	 */
	addch(' ');
	putstr_trunc(((PANEL_MENU *)panel)->menu[ln],display.scrcols - 3,0);
}

static void
line_conf(int ln)
{
	addch(' ');
	putstr_trunc(panel_config.config[ln].var,CFGVAR_MAXLEN,0);
	addstr(" = ");
	putstr_trunc(config_print_value(ln),
	  display.scrcols - CFGVAR_MAXLEN - 6,0);
}

static void
line_sort(int ln)
{
	addstr(panel_sort.order == ln ? " [x] " : " [ ] ");
	putstr_trunc(panel_sort.menu[ln],display.scrcols - 7,0);
}

static void
line_completion(int ln)
{
	int width;

	addch(' ');
	width = display.scrcols - 3;
	width -= putstr_trunc(USTR(panel_compl.candidate[ln].name),
	  width,OPT_NOPAD);
	if (width == 0)
		return;
	/* no leading tilde and no trailing slash for COMPL_TYPE_USER */
	if (panel_compl.candidate[ln].isdir &&
	  panel_compl.compl_type != COMPL_TYPE_USER) {
		addch('/');
		width--;
	}
	char_line(' ',width);
}

static void
line_help(int ln)
{
	const char *str;

	addch(' ');
	str = panel_help.txt[ln];
	if (str[0] == '@' && str[1] == 'L' && str[2] == '=') {
		/* it's a link */
		for (str += 3; *str++; )
			;
		attron(attrb);
		putstr_trunc(str,display.scrcols - 4,0);
		attroff(attrb);
	}
	else
		putstr_trunc(str,display.scrcols - 4,0);
	addch(' ');
}

static void
one_line(int y)
{
	int ln;
	FLAG current;

	move(2 + y,0);
	ln = panel->top + y;
	if (ln >= panel->cnt) {
		clrtoeol();
		return;
	}

	current = panel->curs == ln;
	if (current) {
		addch('>');
		if (panel->type != PANEL_TYPE_HELP)
			attrset(attrr);
	}
	else
		addch(' ');
	if (ln < 0)
		putstr_trunc(" --- (leave this panel) ---",
		  display.scrcols - 2,0);
	else
		(*put_line)(ln);
	if (current) {
		attrset(A_NORMAL);
		addch('<');
	}
	else 
		addch(' ');
}

static void
draw_panel(int opt_enable)
{
	static FLAG need_wininfo;
	static int save_top, save_curs, save_ptype = PANEL_TYPE_NONE;
	int y;

	/* special case: panel is empty */
	if (panel->cnt == 0) {
		if (save_ptype != PANEL_TYPE_EMPTY) {
			for (y = 0; y < display.panlines; y++) {
				move(2 + y,0);
				clrtoeol();
			}
			win_position(MSG_NODATA);
			win_info();
		}
		save_ptype = PANEL_TYPE_EMPTY;
		return;
	}

	if (panel->type != save_ptype) {
		/* panel type has changed */
		need_wininfo = 0;
		opt_enable = 0;
		switch (save_ptype = panel->type) {
			case PANEL_TYPE_COMPL:
				need_wininfo = 1;
				put_line = line_completion;
				break;
			case PANEL_TYPE_CONFIG:
				need_wininfo = 1;
				put_line = line_conf;
				break;
			case PANEL_TYPE_DIR_BROWSE:
				put_line = line_dir_browse;
				break;
			case PANEL_TYPE_DIR_SPLIT:
				put_line = line_dir_split;
				break;
			case PANEL_TYPE_FILE:
				need_wininfo = 1;
				put_line = line_file;
				break;
			case PANEL_TYPE_HELP:
				put_line = line_help;
				break;
			case PANEL_TYPE_HIST:
				put_line = line_hist;
				break;
			case PANEL_TYPE_LOCATE:
				need_wininfo = 1;
				put_line = line_locate;
				break;
			case PANEL_TYPE_MENU:
				put_line = line_menu;
				break;
			case PANEL_TYPE_SORT:
				put_line = line_sort;
				break;
		}
		if (!need_wininfo)
			/*
			 * this is not a typo !
			 * run win_info() just once to erase the old info line
			 */
			win_info();
	}

	if (opt_enable && save_top == panel->top) {
		/* redraw old and new lines */
		one_line(save_curs - panel->top);
		if (save_curs != panel->curs)
			one_line(panel->curs - panel->top);
	}
	else {
		/* redraw all lines */
		for (y = 0; y < display.panlines; y++)
			one_line(y);
		save_top = panel->top;
	}
	save_curs = panel->curs;

	win_position(MSG_NORMAL);
	if (need_wininfo)
		win_info();
}

/* win_panel() without optimization */
void
win_panel(void)
{
	draw_panel(0);
}

/*
 * win_panel() with optimization
 *
 * use this win_panel() version if the only change made since last
 * win_panel() call is a cursor movement or a modification of the
 * current line
 */
void
win_panel_opt(void)
{
	draw_panel(1);
}
